#include <math.h>
#include <propeller.h>
#include <sys/thread.h>
//**********************************//
#include "Motor_led_Driver.h"
#include "MPU6050_XS.h"
#include "FullDuplexSerial.h"
#include "sensfusion6.h"
#include "pid.h"
#include "controller.h"
#include "Nordic_Demo.h"
//***********************************//
#define ATTITUDE_UPDATE_RATE_DIVIDER  2
#define FUSION_UPDATE_DT  (float)(1.0 / (IMU_UPDATE_FREQ / ATTITUDE_UPDATE_RATE_DIVIDER))
#define MOTORS_PWM_BITS     9
#define MOTORS_PWM_PERIOD   ((1<<MOTORS_PWM_BITS) - 1)
#define MOTORS_PWM_PRESCALE 0
#define C_BITS_TO_16(X) ((X)<<(16-MOTORS_PWM_BITS))
#define C_16_TO_BITS(X) ((X)>>(16-MOTORS_PWM_BITS)&((1<<MOTORS_PWM_BITS)-1))
#define STACK_SIZE 20
//#define _DEBUG
//***********************************//
static Axis3f gyro; // Gyro axis data in deg/s
static Axis3f acc;  // Accelerometer axis data in m
static float eulerRollActual;
static float eulerPitchActual;
static float eulerYawActual;
static float eulerRollDesired;
static float eulerPitchDesired;
static float eulerYawDesired;
static float rollRateDesired;
static float pitchRateDesired;
static float yawRateDesired;
static float fusionDt;
static int cog3_stack[STACK_SIZE];
static _thread_state_t thread_data;
//**************************************//
typedef enum
{
  RATE,
  ANGLE
} RPYType;

RPYType rollType;
RPYType pitchType;
RPYType yawType;
//**************************************//
int loopTime;
int timer;
uint8_t  preflight_check;
uint8_t  attitudeCounter = 0;
int16_t  actuatorRoll;
int16_t  actuatorPitch;
int16_t  actuatorYaw;
uint16_t actuatorThrust;
//**************************************//
       void     pm_failsafe_loop(void);
       void     Pth_stabilizer_loop(void);
static void     distributePower(const uint16_t thrust, const int16_t roll, const int16_t pitch, const int16_t yaw);
static uint16_t limitThrust(int32_t value);
       uint8_t  Preeflight(void);
//**************************************//

int main(void)
{   
    waitcnt(CNT+CLKFREQ);
    // Start Full Duplex serial if later in Debug Mode
    #ifdef _DEBUG
    FullDuplexSerial_Start( 31,  30,  0,  115200);
    FullDuplexSerial_Str("H6 UAV Demo Software V0.0.1\r");
    #endif
    if (Preeflight()==0){
        #ifdef _DEBUG
        FullDuplexSerial_Str("Start Stabilizer COG 5\r");
        #endif
        // From here start cog 3 with loop  for the stabilizer
       _start_cog_thread(cog3_stack + STACK_SIZE, Pth_stabilizer_loop, NULL, &thread_data);
        //Pth_stabilizer_loop();
    }
    else
    {
    //TODO: Errror Handling
    #ifdef _DEBUG 
    FullDuplexSerial_Str("Preflight Check Failed\r");
    #endif
    }
    
    //Nordic_Demo_Loop();
    #ifdef _DEBUG
    FullDuplexSerial_Str("COG 0 PM and Failsafe LOOP\r");
    #endif
    pm_failsafe_loop();   
    
    // Should never get here 
    return ; 
}

uint8_t Preeflight(void){
    #ifdef _DEBUG
    FullDuplexSerial_Str("Setup Preflight\r");
    #endif
    // Setup stuff we need to get of the ground
    rollRateDesired = 0;
    pitchRateDesired = 0;
    yawRateDesired = 0;
    actuatorThrust = 10000;
    
    #ifdef _DEBUG
    FullDuplexSerial_Str("Start Motor and LED Driver COG 1\r");
    #endif
    // Start Motor Driver in cog nr 2      
    Motor_led_Driver_Start();
    #ifdef _DEBUG
    FullDuplexSerial_Str("Start IMU Driver COG 2\r");
    #endif
    // Start IMU/I2C Driver in cog nr 3
    MPU6050_XS_Start(17,16,21,20,0);
    #ifdef _DEBUG
    FullDuplexSerial_Str("Start PID Driver\r");
    #endif
    // Setup The controller PIDS
    controllerInit();
    #ifdef _DEBUG
    FullDuplexSerial_Str("Start Nordic Driver COG 3&4 \r");
    #endif
    Nordic_Demo_Demo();
    // Setup RC Reciver PPM
    
    
return preflight_check;
}

void Pth_stabilizer_loop(void){
    

    while(1){
      timer = -CNT;
        //ToDo : make time counter for loop
        MPU6050_XS_Imu6Read(&gyro, &acc);
      if (++attitudeCounter >= ATTITUDE_UPDATE_RATE_DIVIDER)
      {
        sensfusion6UpdateQ(gyro.x, gyro.y, gyro.z, acc.x, acc.y, acc.z, FUSION_UPDATE_DT);

        sensfusion6GetEulerRPY(&eulerRollActual, &eulerPitchActual, &eulerYawActual);
        controllerCorrectAttitudePID(eulerRollActual, eulerPitchActual, eulerYawActual, eulerRollDesired, eulerPitchDesired, -eulerYawDesired, &rollRateDesired, &pitchRateDesired, &yawRateDesired);
        attitudeCounter = 0;
      }
      if (rollType == RATE)
      {
        rollRateDesired = eulerRollDesired;
      }
      if (pitchType == RATE)
      {
        pitchRateDesired = eulerPitchDesired;
      }
      if (yawType == RATE)
      {
        yawRateDesired = -eulerYawDesired;
      }

      // TODO: Investigate possibility to subtract gyro drift.
      controllerCorrectRatePID(gyro.x, -gyro.y, gyro.z, rollRateDesired, pitchRateDesired, yawRateDesired);
      controllerGetActuatorOutput(&actuatorRoll, &actuatorPitch, &actuatorYaw);
         
      //Update 6 Motors       
      distributePower(actuatorThrust,actuatorPitch, actuatorRoll,  -actuatorYaw);          
        
        loopTime = (timer+CNT-16)/80000;
  
    }
}

static void distributePower(const uint16_t thrust, const int16_t roll, const int16_t pitch, const int16_t yaw)
{
  uint16_t motorFrontLeft;
  uint16_t motorFrontRight;
  uint16_t motorMiddleLeft;
  uint16_t motorMiddleRight;
  uint16_t motorRearLeft;
  uint16_t motorRearRight;

  motorFrontLeft =      limitThrust(thrust + roll - pitch + yaw);
  motorFrontRight =     limitThrust(thrust - roll - pitch - yaw);
  motorMiddleLeft =     limitThrust(thrust + roll - yaw);
  motorMiddleRight =    limitThrust(thrust - roll + yaw);
  motorRearLeft =       limitThrust(thrust + roll + pitch + yaw);
  motorRearRight =      limitThrust(thrust - roll + pitch - yaw);
        
  Motor_led_Driver_Pwm(13,C_16_TO_BITS(motorFrontLeft),511);
  Motor_led_Driver_Pwm(12,C_16_TO_BITS(motorFrontRight),511);
  Motor_led_Driver_Pwm(11,C_16_TO_BITS(motorMiddleLeft),511);
  Motor_led_Driver_Pwm(10,C_16_TO_BITS(motorMiddleRight),511);
  Motor_led_Driver_Pwm(9 ,C_16_TO_BITS(motorRearLeft),511);
  Motor_led_Driver_Pwm(8 ,C_16_TO_BITS(motorRearRight),511);
}

static uint16_t limitThrust(int32_t value)
{
  if(value > UINT16_MAX)
  {
    value = UINT16_MAX;
  }
  else if(value < 0)
  {
    value = 0;
  }

  return (uint16_t)value;
}

void pm_failsafe_loop(void){
    // Power Management loop cog 0 incl failsafe
    while(1){
    
        // Debug print out
        #ifdef _DEBUG
        waitcnt(CNT+CLKFREQ/100);
        FullDuplexSerial_Str("Loop time ms --");
        FullDuplexSerial_Dec(loopTime); 
        FullDuplexSerial_Str("\n"); 
        #endif
        // Power Management
        // 1 check ADC
        // 2 if below min start autodecent 
        
        // Fail Safe RC Reciver
        // 1 check if we have rssi
        // if not go to hover pitch,roll,yaw = 0
         
    }
}